#ifndef _LIR_H_
#define _LIR_H_ 
#include <cassert>
#include <iostream>
#include <list>
#include <map>
#include <set>
#include <sstream>
#include <string>
#include <utility>
#include <vector>
#include <memory>
#include <stdexcept>
#include "../IR/llvm.h"
class LIROperand;
class LIRConstant;
class LIRVirtReg;
class LIRPhysReg;
class LIRInstruction;
class LIRIntInst;
class LIRFloatInst;
class LIRVmrsInst;
class LIRLdrInst;
class LIRVldrInst;
class LIRStrInst;
class LIRVstrInst;
class LIRBInst;
class LIRBlInst;
class LIRPushInst;
class LIRPopInst;
class LIRDataDef;
class LIRLabel;
class LIRGlobalVariable;
class LIRBasicBlock;
class LIRFunction;

struct LIRUse {
    LIRInstruction* val_;
    int arg_no_;
    LIRUse(LIRInstruction* val, int no) : val_(val), arg_no_(no) {}
};
class LIRDataDef {
public:
    enum LIRDataDefTy { LWordTy, LZeroTy };
    explicit LIRDataDef(LIRDataDefTy tid, unsigned expr) : tid_(tid), expr_(expr) {}
    ~LIRDataDef() = default;
    std::string print();
    LIRDataDefTy tid_;
    unsigned expr_;
};
class LIROperand {
public:
    enum LIROprTy {
       LOVirtRegTy,LOVirtFRegTy,LOPhysRegTy,LOPhysFRegTy,LOConstantTy,LOShiftTy,LFunctionTy,LBasicBlockTy,LGlobalVariableTy
    };
    LIROprTy tid_;
    std::list<LIRUse> use_list_;
    explicit LIROperand(LIROprTy tid) : tid_(tid) {}
    ~LIROperand() = default;
    virtual std::string print() = 0;
    std::list<LIRUse>::iterator add_use(LIRInstruction* val, int arg_no) {
        use_list_.emplace_back(val, arg_no);
        std::list<LIRUse>::iterator re = use_list_.end();
        return --re;
    }
    void remove_use(std::list<LIRUse>::iterator it) {
        use_list_.erase(it);
    }
   
};
class LIRConstant : public LIROperand {
public:
    int value_;
    LIRConstant(int val) : LIROperand(LOConstantTy), value_(val) {}
    ~LIRConstant() = default;
    virtual std::string print() override;
   
};
class LIRShift : public LIROperand {
public:
    enum LIRShiftTy { LSAsrTy, LSLslTy, LSLsrTy, LSRorTy };
    LIRShift(LIRShiftTy tid, int amount) : LIROperand(LOShiftTy), tid_(tid), amount_(amount) {}
    ~LIRShift() = default;
    const static std::map<LIRShiftTy, std::string> name_;
    virtual std::string print() override;
    LIRShiftTy tid_;
    int amount_;
};
class LIRVirtReg : public LIROperand {
public:
    LIRVirtReg(int rid) : LIROperand(LOVirtRegTy), rid_(rid) {}
    ~LIRVirtReg() = default;
    virtual std::string print() override;
    int rid_;
};
class LIRPhysReg : public LIROperand {
public:
    LIRPhysReg(int rid) : LIROperand(LOPhysRegTy), rid_(rid) {}
    ~LIRPhysReg() = default;
    const static std::map<int, std::string> name_;
    virtual std::string print() override;
    int rid_;
};

class LIRVirtFReg : public LIROperand {
public:
    LIRVirtFReg(int frid) : LIROperand(LOVirtFRegTy), frid_(frid) {}
    ~LIRVirtFReg() = default;
    virtual std::string print() override;
    int frid_;
};
class LIRPhysFReg : public LIROperand {
public:
    LIRPhysFReg(int frid) : LIROperand(LOPhysFRegTy), frid_(frid) {}
    ~LIRPhysFReg() = default;
    const static std::map<int, std::string> name_;
    virtual std::string print() override;
    int frid_;
};

class LIRModule {
public:
    explicit LIRModule() {
        total_rid_ = 0;
        total_frid_ = 0;
    }
    ~LIRModule() {}
    virtual std::string print();
    void add_global_variable(LIRGlobalVariable* g) { global_list_.push_back(g); }
    void add_function(LIRFunction* f) { function_list_.push_back(f); }
    LIRConstant* get_LIRConstant(int val) ;
    LIRVirtReg* get_LIRVirtReg(int rid) ;
    LIRVirtReg* get_LIRVirtReg() ;
    LIRPhysReg* get_LIRPhysReg(int rid) ;
    LIRVirtFReg* get_LIRVirtFReg(int frid) ;
    LIRVirtFReg* get_LIRVirtFReg() ;
    LIRPhysFReg* get_LIRPhysFReg(int frid) ;
    LIRDataDef* get_LIRDataDef(LIRDataDef::LIRDataDefTy tid, int expr) ;
    LIRShift* get_LIRShift(LIRShift::LIRShiftTy tid, int amount) ;   
   
    std::vector<LIRGlobalVariable*> global_list_;
    std::vector<LIRFunction*> function_list_;
    std::map<int, LIRConstant*> LIRConstant_map_;
    std::map<int, LIRVirtReg*> LIRVirtReg_map_;
    std::map<int, LIRPhysReg*> LIRPhysReg_map_;
    std::map<int, LIRVirtFReg*> LIRVirtFReg_map_;
    std::map<int, LIRPhysFReg*> LIRPhysFReg_map_;
    std::map<std::pair<LIRShift::LIRShiftTy, int>, LIRShift*> LIRShift_map_;
    std::map<std::pair<LIRDataDef::LIRDataDefTy, int>, LIRDataDef*> LIRDataDef_map_;
    int total_rid_;
    int total_frid_;
};
class LIRLabel : public LIROperand {
public:
    LIRLabel(LIROprTy tid, const std::string& name) : LIROperand(tid), name_(name) {}
    ~LIRLabel() = default;
    virtual std::string print() = 0;
    std::string name_;
};
class LIRGlobalVariable : public LIRLabel {
public:
    enum LIRDataSectionTy { LDataTy, LBssTy };
    LIRGlobalVariable(std::string name, LIRModule* m, LIRDataSectionTy tid) : LIRLabel(LGlobalVariableTy, name), tid_(tid), parent_(m) { m->add_global_variable(this); }
    void add_data_def(LIRDataDef* dd) { init_val_.push_back(dd); }
    virtual std::string print() override;
    LIRDataSectionTy tid_;
    LIRModule* parent_;
    std::vector<LIRDataDef*> init_val_;
};
class LIRFunction : public LIRLabel {
public:
    LIRFunction(const std::string& name, LIRModule* m, int has_result, int num_args) : LIRLabel(LFunctionTy, name), parent_(m), has_result_(has_result), num_args_(num_args) { parent_->add_function(this); }
    LIRFunction(const std::string& name, int has_result, int num_args) : LIRLabel(LFunctionTy, name), parent_(nullptr), has_result_(has_result), num_args_(num_args) { }
    ~LIRFunction();
    virtual std::string print() override;
    void add_basic_block(LIRBasicBlock* bb) { basic_blocks_.push_back(bb); }
    std::vector<LIRBasicBlock*> basic_blocks_;
    int has_result_;
    int num_args_;
    LIRModule* parent_;
    std::map<void*, int> var_offset_;
    bool is_libfunc(){
        if (name_ == "putint" || name_ == "putch" || name_ == "putarray" || name_ == "_sysy_starttime" || name_ == "_sysy_stoptime" || name_ == "__aeabi_memclr4" || name_ == "__aeabi_memset4" || name_ == "__aeabi_memcpy4"
        ||name_ == "getint" || name_ == "getch" || name_ == "getarray"
        ||name_ == "getfloat" ||name_=="getfarray"||name_ == "putfloat" || name_ == "putfarray"){
            return true;
        }else{
            return false;
        }
    }
    int stackframe_args_size_=0;
    std::vector<LIROperand*> stackframe_saveregs_;
    std::vector<std::pair<AllocaInst*,int>> stackframe_allocas_;
    std::vector<LIROperand*> stackframe_spills_;
    std::map<AllocaInst*, int> allocas_offset_;
    int allocas_offset_end_=0;
    std::vector<int> temps_offset_;
    int temps_offset_end_=0;
    std::map<LIROperand*, int> spills_offset_;
    int spills_offset_end_=0;
    int stackframe_offset_end_=0;
    int saveregs_offset_end_=0;
    int args_offset_end_=0;
    std::map<LIRInstruction*,int> args_offset_;
    std::vector<LIROperand*> args_passmethod_;
};
class LIRBasicBlock : public LIRLabel {
public:
    explicit LIRBasicBlock(const std::string& name, LIRFunction* f) : LIRLabel(LBasicBlockTy, ".LBB."+f->name_+"."+name), parent_(f) { f->add_basic_block(this); }
    void add_inst_back(LIRInstruction* instr) ;
    void add_pre_basic_block(LIRBasicBlock* bb) { pre_bbs_.push_back(bb); }
    void add_succ_basic_block(LIRBasicBlock* bb) { succ_bbs_.push_back(bb); }
    void delete_instr(LIRInstruction* instr);
    void replace_instr(LIRInstruction* oldinst,LIRInstruction* newinst);
    void add_inst_front(LIRInstruction* instr);
    bool add_inst_before_br(LIRInstruction* instr);
    bool add_inst_before(LIRInstruction *instr, LIRInstruction *dst);
    bool add_inst_after(LIRInstruction *instr, LIRInstruction *dst);
    virtual std::string print() override;
    std::list<LIRInstruction*> instr_list_;
    LIRFunction* parent_;
    int bb_idx_;
    std::vector<LIRBasicBlock*> pre_bbs_;
    std::vector<LIRBasicBlock*> succ_bbs_;
    int incomingForwardBranches = 0;
    int loopIdx = 0;
    int loopDepth = 0;
    int getLoopIndex() const { return loopIdx; }
    int getLoopDepth() const { return loopDepth; }
    void setLoopIndex(int idx) { loopIdx = idx; }
    void setLoopDepth(int depth) { loopDepth = depth; }
    void setIncomingForwardBranches(int num) { incomingForwardBranches = num; }
    void decreaseForwardBranches() { incomingForwardBranches -= 1; }
    bool isForwardBranchesEQZero() const { return incomingForwardBranches == 0; }
};
class LIRInstruction {
public:
    enum LIRInstTy { LAddTy, LSubTy,LRsbTy, LMulTy, LMlaTy,LMlsTy,LSdivTy,LUdivTy,LAndTy,LOrrTy,LEorTy, LLslTy, LLsrTy, LAsrTy,LRorTy, LBTy,LBlTy,LBxTy, LMovTy,LVmovTy,LVmovf32Ty,LMov32Ty, LLdrTy,LVldrTy, LStrTy,LVstrTy, LPushTy, LPopTy,LVpushTy,LVpopTy, LCmpTy,LVcmpTy,LVaddTy,LVsubTy,LVdivTy,LVmulTy,LVmlaTy,LVmlsTy,LVnegTy,LVcvtFSTy,LVcvtSFTy,LVmrsTy };
    enum LIRCondTy { LCondGT, LCondGE, LCondLT, LCondLE, LCondHI, LCondCS, LCondCC, LCondLS, LCondEQ, LCondNE, LCondVC, LCondVS, LCondNone };
    enum LIRAddrTy { LRegisterOffsetTy, LPreIndexedTy, LPostIndexedTy };
    const static std::map<LIRInstTy, std::string> inst_name_;
    const static std::map<LIRCondTy, std::string> cond_name_;
    LIRInstruction(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, int num_ops, LIRBasicBlock* bb) : op_id_(op_id), cond_id_(cond_id), result_(result), num_ops_(num_ops), parent_(bb) {
        operands_.resize(num_ops_, nullptr);
        opr_use_pos_.resize(num_ops_);
        parent_->add_inst_back(this);
        if (result != nullptr) {
            res_use_pos_ = result->add_use(this, -1);
        }
    }
    LIRInstruction(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, int num_ops) : op_id_(op_id), cond_id_(cond_id), result_(result), num_ops_(num_ops), parent_(nullptr) {
        operands_.resize(num_ops_, nullptr);
        opr_use_pos_.resize(num_ops_);
        if (result != nullptr) {
            res_use_pos_ = result->add_use(this, -1);
        }
    }
    void set_operand(int i, LIROperand* v) ;
    void remove_operand(int i) ;
    void remove_operand_real(int i) ;
    void set_result(LIROperand* v) ;
    void remove_result() ;
    void remove_result_and_operands() ;
    std::vector<LIROperand*>get_operands_and_result();
    LIROperand* get_operand(int i) const { return operands_[i]; }
    virtual std::string print() = 0;
    LIRBasicBlock* parent_;
    LIRInstTy op_id_;
    LIRCondTy cond_id_;
    int num_ops_;
    LIROperand* result_;
    std::vector<LIROperand*> operands_;
    std::list<LIRUse>::iterator res_use_pos_;
    std::vector<std::list<LIRUse>::iterator> opr_use_pos_;
    std::vector<std::map<LIRInstruction*,int>::iterator> args_offset_use_pos_;
};
class LIRIntInst : public LIRInstruction {
public:
    LIRIntInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2, LIRBasicBlock* bb) : LIRInstruction(op_id, cond_id, result, 2, bb) {
        set_operand(0, v1);
        set_operand(1, v2);
    }
    LIRIntInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2) : LIRInstruction(op_id, cond_id, result, 2) {
        set_operand(0, v1);
        set_operand(1, v2);
    }
    LIRIntInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIRBasicBlock* bb) : LIRInstruction(op_id, cond_id, result, 1, bb) { set_operand(0, v1); }
    LIRIntInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1) : LIRInstruction(op_id, cond_id, result, 1) { set_operand(0, v1); }
    LIRIntInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2, LIROperand* v3, LIRBasicBlock* bb) : LIRInstruction(op_id, cond_id, result, 3, bb) {
        set_operand(0, v1);
        set_operand(1, v2);
        set_operand(2, v3);
    }
    LIRIntInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2, LIROperand* v3) : LIRInstruction(op_id, cond_id, result, 3) {
        set_operand(0, v1);
        set_operand(1, v2);
        set_operand(2, v3);
    }
    virtual std::string print() override;
};
class LIRFloatInst : public LIRInstruction {
public:
    LIRFloatInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2, LIRBasicBlock* bb) : LIRInstruction(op_id, cond_id, result, 2, bb) {
        set_operand(0, v1);
        set_operand(1, v2);
    }
    LIRFloatInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2) : LIRInstruction(op_id, cond_id, result, 2) {
        set_operand(0, v1);
        set_operand(1, v2);
    }
    LIRFloatInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIRBasicBlock* bb) : LIRInstruction(op_id, cond_id, result, 1, bb) { set_operand(0, v1); }
    LIRFloatInst(LIRInstTy op_id, LIRCondTy cond_id, LIROperand* result, LIROperand* v1) : LIRInstruction(op_id, cond_id, result, 1) { set_operand(0, v1); }
    virtual std::string print() override;
};
class LIRMovInst : public LIRInstruction {
public:
    LIRMovInst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIRBasicBlock* bb) : LIRInstruction(LMovTy, cond_id, result, 1, bb) { set_operand(0, v1); }
    LIRMovInst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1) : LIRInstruction(LMovTy, cond_id, result, 1) { set_operand(0, v1); }
    virtual std::string print() override;
};
class LIRVmovInst : public LIRInstruction {
public:
    LIRVmovInst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIRBasicBlock* bb) : LIRInstruction(LVmovTy, cond_id, result, 1, bb) { set_operand(0, v1); }
    LIRVmovInst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1) : LIRInstruction(LVmovTy, cond_id, result, 1) { set_operand(0, v1); }
    virtual std::string print() override;
};
class LIRVmovf32Inst : public LIRInstruction {
public:
    LIRVmovf32Inst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIRBasicBlock* bb) : LIRInstruction(LVmovf32Ty, cond_id, result, 1, bb) { set_operand(0, v1); }
    LIRVmovf32Inst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1) : LIRInstruction(LVmovf32Ty, cond_id, result, 1) { set_operand(0, v1); }
    virtual std::string print() override;
};
class LIRMov32Inst : public LIRInstruction {
public:
    LIRMov32Inst(LIRCondTy cond_id, LIROperand* result, LIRConstant* v1, LIRBasicBlock* bb) : LIRInstruction(LMov32Ty, cond_id, result, 1, bb){ set_operand(0, v1); }
    LIRMov32Inst(LIRCondTy cond_id, LIROperand* result, LIRConstant* v1) : LIRInstruction(LMov32Ty, cond_id, result, 1){ set_operand(0, v1); }
    LIRMov32Inst(LIRCondTy cond_id, LIROperand* result, LIRFunction* v1, LIRBasicBlock* bb) : LIRInstruction(LMov32Ty, cond_id, result, 1, bb) { set_operand(0, v1); }
    LIRMov32Inst(LIRCondTy cond_id, LIROperand* result, LIRGlobalVariable* v1, LIRBasicBlock* bb) : LIRInstruction(LMov32Ty, cond_id, result, 1, bb) { set_operand(0, v1); }
    virtual std::string print() override;
};
class LIRLdrInst : public LIRInstruction {
public:
    LIRLdrInst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2, LIRAddrTy addr_tid, LIRBasicBlock* bb) : LIRInstruction(LLdrTy, cond_id, result, 2, bb), addr_tid_(addr_tid) {
        set_operand(0, v1);
        set_operand(1, v2);
    }
    LIRLdrInst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2, LIRAddrTy addr_tid) : LIRInstruction(LLdrTy, cond_id, result, 2), addr_tid_(addr_tid) {
        set_operand(0, v1);
        set_operand(1, v2);
    }
    virtual std::string print() override;
    LIRAddrTy addr_tid_;
};
class LIRVldrInst : public LIRInstruction {
public:
    LIRVldrInst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2, LIRBasicBlock* bb) : LIRInstruction(LVldrTy, cond_id, result, 2, bb) {
        set_operand(0, v1);
        set_operand(1, v2);
    }
    LIRVldrInst(LIRCondTy cond_id, LIROperand* result, LIROperand* v1, LIROperand* v2) : LIRInstruction(LVldrTy, cond_id, result, 2) {
        set_operand(0, v1);
        set_operand(1, v2);
    }
    virtual std::string print() override;
};
class LIRStrInst : public LIRInstruction {
public:
    LIRStrInst(LIRCondTy cond_id, LIROperand* v1, LIROperand* v2, LIROperand* v3, LIRAddrTy addr_tid, LIRBasicBlock* bb) : LIRInstruction(LStrTy, cond_id, nullptr, 3, bb), addr_tid_(addr_tid) {
        set_operand(0, v1);
        set_operand(1, v2);
        set_operand(2, v3);
    }
    LIRStrInst(LIRCondTy cond_id, LIROperand* v1, LIROperand* v2, LIROperand* v3, LIRAddrTy addr_tid) : LIRInstruction(LStrTy, cond_id, nullptr, 3), addr_tid_(addr_tid) {
        set_operand(0, v1);
        set_operand(1, v2);
        set_operand(2, v3);
    }
    virtual std::string print() override;
    LIRAddrTy addr_tid_;
};
class LIRVstrInst : public LIRInstruction {
public:
    LIRVstrInst(LIRCondTy cond_id, LIROperand* v1, LIROperand* v2, LIROperand* v3, LIRBasicBlock* bb) : LIRInstruction(LVstrTy, cond_id, nullptr, 3, bb) {
        set_operand(0, v1);
        set_operand(1, v2);
        set_operand(2, v3);
    }
    LIRVstrInst(LIRCondTy cond_id, LIROperand* v1, LIROperand* v2, LIROperand* v3) : LIRInstruction(LVstrTy, cond_id, nullptr, 3) {
        set_operand(0, v1);
        set_operand(1, v2);
        set_operand(2, v3);
    }
    virtual std::string print() override;
};
class LIRBInst : public LIRInstruction {
public:
    LIRBInst(LIRCondTy cond_id, LIRBasicBlock* target_bb, LIRBasicBlock* bb) : LIRInstruction(LBTy, cond_id, nullptr, 1, bb) {set_operand(0,target_bb);}
    virtual std::string print() override;
};
class LIRBlInst : public LIRInstruction {
public:
    LIRBlInst(LIRCondTy cond_id, LIRFunction* target_func, bool result_unused, LIRBasicBlock* bb) : LIRInstruction(LBlTy, cond_id, nullptr, 1, bb),result_unused_(result_unused) {set_operand(0,target_func);}
    LIRBlInst(LIRCondTy cond_id, LIRFunction* target_func, bool result_unused) : LIRInstruction(LBlTy, cond_id, nullptr, 1),result_unused_(result_unused) {set_operand(0,target_func);}
    virtual std::string print() override;
    bool result_unused_;
};
class LIRBxInst : public LIRInstruction {
public:
    LIRBxInst(LIRCondTy cond_id, LIROperand* v1, LIRBasicBlock* bb) : LIRInstruction(LBxTy, cond_id, nullptr, 1, bb) {set_operand(0,v1);}
    LIRBxInst(LIRCondTy cond_id, LIROperand* v1) : LIRInstruction(LBxTy, cond_id, nullptr, 1) {set_operand(0,v1);}
    virtual std::string print() override;
};
class LIRVmrsInst : public LIRInstruction {
public:
    LIRVmrsInst(LIRBasicBlock* bb) : LIRInstruction(LVmrsTy, LCondNone, nullptr, 0, bb) {}
    virtual std::string print() override;
};
class LIRPushInst : public LIRInstruction {
public:
    LIRPushInst(LIRCondTy cond_id, const std::vector<LIROperand*>& reglist, LIRBasicBlock* bb) : LIRInstruction(LPushTy, cond_id, nullptr, reglist.size(), bb) {
        for (int i = 0; i < reglist.size(); ++i) {
            set_operand(i, reglist[i]);
        }
    }
    LIRPushInst(LIRCondTy cond_id, const std::vector<LIROperand*>& reglist) : LIRInstruction(LPushTy, cond_id, nullptr, reglist.size()) {
        for (int i = 0; i < reglist.size(); ++i) {
            set_operand(i, reglist[i]);
        }
    }
    virtual std::string print() override;
};
class LIRPopInst : public LIRInstruction {
public:
    LIRPopInst(LIRCondTy cond_id, const std::vector<LIROperand*>& reglist, LIRBasicBlock* bb) : LIRInstruction(LPopTy, cond_id, nullptr, reglist.size(), bb) {
        for (int i = 0; i < reglist.size(); ++i) {
            set_operand(i, reglist[i]);
        }
    }
    LIRPopInst(LIRCondTy cond_id, const std::vector<LIROperand*>& reglist) : LIRInstruction(LPopTy, cond_id, nullptr, reglist.size()) {
        for (int i = 0; i < reglist.size(); ++i) {
            set_operand(i, reglist[i]);
        }
    }
    virtual std::string print() override;
};
class LIRVpushInst : public LIRInstruction {
public:
    LIRVpushInst(LIRCondTy cond_id, const std::vector<LIROperand*>& reglist, LIRBasicBlock* bb) : LIRInstruction(LVpushTy, cond_id, nullptr, reglist.size(), bb) {
        for (int i = 0; i < reglist.size(); ++i) {
            set_operand(i, reglist[i]);
        }
    }
    LIRVpushInst(LIRCondTy cond_id, const std::vector<LIROperand*>& reglist) : LIRInstruction(LVpushTy, cond_id, nullptr, reglist.size()) {
        for (int i = 0; i < reglist.size(); ++i) {
            set_operand(i, reglist[i]);
        }
    }
    virtual std::string print() override;
};
class LIRVpopInst : public LIRInstruction {
public:
    LIRVpopInst(LIRCondTy cond_id, const std::vector<LIROperand*>& reglist, LIRBasicBlock* bb) : LIRInstruction(LVpopTy, cond_id, nullptr, reglist.size(), bb) {
        for (int i = 0; i < reglist.size(); ++i) {
            set_operand(i, reglist[i]);
        }
    }
    LIRVpopInst(LIRCondTy cond_id, const std::vector<LIROperand*>& reglist) : LIRInstruction(LVpopTy, cond_id, nullptr, reglist.size()) {
        for (int i = 0; i < reglist.size(); ++i) {
            set_operand(i, reglist[i]);
        }
    }
    virtual std::string print() override;
};
#endif
