#ifndef TEST_AST_H
#define TEST_AST_H
#pragma once

#include <cstring>
#include <vector>
#include <string>
#include <memory>
#include <iostream>


using namespace std;

enum STYPE {
    SEMI, ASS, EXP, CONT, BRE, RET, BLK, SEL, ITER
};

enum UOP {
    UOP_ADD, UOP_MINUS, UOP_NOT
};

enum AOP {
    AOP_ADD, AOP_MINUS
};

enum MOP {
    MOP_MUL, MOP_DIV, MOP_MOD
};

enum EOP {
    EOP_EQ, EOP_NEQ
};

enum ROP {
    ROP_GTE, ROP_LTE, ROP_GT, ROP_LT
};



enum TYPE {
    TYPE_VOID, TYPE_INT, TYPE_FLOAT
};

class BaseAST;

class COMPUnitAST;
class DECLDefAST;
class DECLAST;
class DefListAST;
class DEFAST;
class ArraysAST;
class InitValListAST;
class INITValAST;
class FUNCDefAST;
class FuncFParamListAST;
class FUNCFParamAST;
class BLOCKAST;
class BlockItemListAST;
class BLOCKItemAST;
class STMTAST;
class RETURNStmtAST;
class SELECTStmtAST;
class ITERATIONStmtAST;
class LValAST;
class PRIMARYExpAST;
class NUMBERAST;
class UNARYExpAST;
class CALLAST;
class FuncCParamListAST;
class MULExpAST;
class ADDExpAST;
class RELExpAST;
class EQExpAST;
class LANDExpAST;
class LORExpAST;

class Visitor;


class BaseAST {
public:
    virtual void ACcept(Visitor &visitor) = 0;
    BaseAST() = default;
    virtual ~BaseAST() = default;
};

class COMPUnitAST : public BaseAST {
public:
    vector<unique_ptr<DECLDefAST>> declDefList;
    void ACcept(Visitor &visitor)  ;

};

class DECLDefAST : public BaseAST {
public:
    unique_ptr<DECLAST> Decl = nullptr;
    unique_ptr<FUNCDefAST> funcDef = nullptr;
    void ACcept(Visitor &visitor)  ;
};

class DECLAST : public BaseAST {
public:
    TYPE bType;
    bool isConst;
    vector<unique_ptr<DEFAST>> defList;
    void ACcept(Visitor &visitor)  ;
};

class DefListAST {
public:
    vector<unique_ptr<DEFAST>> list;
};

class DEFAST : public BaseAST {
public:
    unique_ptr<string> id;
    vector<unique_ptr<ADDExpAST>> arrays;
    unique_ptr<INITValAST> initVal;
    void ACcept(Visitor &visitor)  ;
};

class ArraysAST {
public:
    vector<unique_ptr<ADDExpAST>> list;
};

class INITValAST:public BaseAST {
public:
    unique_ptr<ADDExpAST> exp;
    vector<unique_ptr<INITValAST>> initValList;
    void ACcept(Visitor &visitor)  ;
};


class InitValListAST {
public:
    vector<unique_ptr<INITValAST>> list;
};

class FUNCDefAST : public BaseAST {
public:
    TYPE funcType;
    unique_ptr<string> id;
    vector<unique_ptr<FUNCFParamAST>> funcFParamList;
    unique_ptr<BLOCKAST> block = nullptr;
    void ACcept(Visitor &visitor)  ;
};

class FuncFParamListAST {
public:
    vector<unique_ptr<FUNCFParamAST>> list;
};

class FUNCFParamAST:public BaseAST {
public:
    TYPE bType;
    unique_ptr<string> id;
    bool isArray = false; 
    vector<unique_ptr<ADDExpAST>> arrays;
    void ACcept(Visitor &visitor)  ;
};

class BLOCKAST : public BaseAST {
public:
    vector<unique_ptr<BLOCKItemAST>> blockItemList;
    void ACcept(Visitor &visitor)  ;
};

class BlockItemListAST {
public:
    vector<unique_ptr<BLOCKItemAST>> list;
};

class BLOCKItemAST : public BaseAST {
public:
    unique_ptr<DECLAST> decl = nullptr;
    unique_ptr<STMTAST> stmt = nullptr;
    void ACcept(Visitor &visitor)  ;
};

class STMTAST : public BaseAST {
public:
    STYPE sType;
    unique_ptr<LValAST> lVal = nullptr;
    unique_ptr<ADDExpAST> exp= nullptr;
    unique_ptr<RETURNStmtAST> returnStmt = nullptr;
    unique_ptr<SELECTStmtAST> selectStmt = nullptr;
    unique_ptr<ITERATIONStmtAST> iterationStmt = nullptr;
    unique_ptr<BLOCKAST> block = nullptr;
    void ACcept(Visitor &visitor)  ;
};

class RETURNStmtAST : public BaseAST {
public:
    unique_ptr<ADDExpAST> exp = nullptr;
    void ACcept(Visitor &visitor)  ;
};

class SELECTStmtAST:public BaseAST {
public:
    unique_ptr<LORExpAST> cond;
    unique_ptr<STMTAST> ifStmt, elseStmt;
    void ACcept(Visitor &visitor)  ;
};

class ITERATIONStmtAST:public BaseAST {
public:
    unique_ptr<LORExpAST> cond;
    unique_ptr<STMTAST> stmt;
    void ACcept(Visitor &visitor)  ;
};

class ADDExpAST : public BaseAST {
public:
    unique_ptr<ADDExpAST> addExp;
    unique_ptr<MULExpAST> mulExp;
    AOP op;
    void ACcept(Visitor &visitor)  ;
};

class MULExpAST :public BaseAST {
public:
    unique_ptr<UNARYExpAST> unaryExp;
    unique_ptr<MULExpAST> mulExp;
    MOP op;
    void ACcept(Visitor &visitor)  ;
};

class UNARYExpAST :public BaseAST {
public:
    unique_ptr<PRIMARYExpAST> primaryExp;
    unique_ptr<CALLAST> call;
    unique_ptr<UNARYExpAST> unaryExp;
    UOP op;
    void ACcept(Visitor &visitor)  ;
};

class PRIMARYExpAST:public BaseAST {
public:
    unique_ptr<ADDExpAST> exp;
    unique_ptr<LValAST> lval;
    unique_ptr<NUMBERAST> number;
    void ACcept(Visitor &visitor)  ;
};

class NUMBERAST:public BaseAST {
public:
    bool isInt;
    union {
        int intval;
        float floatval;
    };
    void ACcept(Visitor &visitor)  ;
};

class LValAST:public BaseAST {
public:
    unique_ptr<string> id;
    vector<unique_ptr<ADDExpAST>> arrays;
    void ACcept(Visitor &visitor)  ;
};

class CALLAST:public BaseAST {
public:
    unique_ptr<string> id;
    vector<unique_ptr<ADDExpAST>> funcCParamList;
    void ACcept(Visitor &visitor)  ;
};

class FuncCParamListAST {
public:
    vector<unique_ptr<ADDExpAST>> list;
};

class RELExpAST:public BaseAST {
public:
    unique_ptr<ADDExpAST> addExp;
    unique_ptr<RELExpAST> relExp;
    ROP op;
    void ACcept(Visitor &visitor) ;
};

class EQExpAST:public BaseAST {
public:
    unique_ptr<RELExpAST> relExp;
    unique_ptr<EQExpAST> eqExp;
    EOP op;
    void ACcept(Visitor &visitor)  ;
};

class LANDExpAST : public BaseAST {
public:
    
    unique_ptr<EQExpAST> eqExp;
    unique_ptr<LANDExpAST> lAndExp;
    void ACcept(Visitor &visitor)  ;
};

class LORExpAST:public BaseAST {
public:
    unique_ptr<LORExpAST> lOrExp;
    unique_ptr<LANDExpAST> lAndExp;
    void ACcept(Visitor &visitor)  ;
};

class Visitor {
public:
    virtual void visit(COMPUnitAST& ast) = 0;
    virtual void visit(DECLDefAST& ast) = 0;
    virtual void visit(DECLAST& ast) = 0;
    virtual void visit(DEFAST& ast) = 0;
    virtual void visit(INITValAST& ast) = 0;
    virtual void visit(FUNCDefAST& ast) = 0;
    virtual void visit(FUNCFParamAST& ast) = 0;
    virtual void visit(BLOCKAST& ast) = 0;
    virtual void visit(BLOCKItemAST& ast) = 0;
    virtual void visit(STMTAST& ast) = 0;
    virtual void visit(RETURNStmtAST& ast) = 0;
    virtual void visit(SELECTStmtAST& ast) = 0;
    virtual void visit(ITERATIONStmtAST& ast) = 0;
    virtual void visit(ADDExpAST& ast) = 0;
    virtual void visit(MULExpAST& ast) = 0;
    virtual void visit(UNARYExpAST& ast) = 0;
    virtual void visit(PRIMARYExpAST& ast) = 0;
    virtual void visit(LValAST& ast) = 0;
    virtual void visit(NUMBERAST& ast) = 0;
    virtual void visit(CALLAST& ast) = 0;
    virtual void visit(RELExpAST& ast) = 0;
    virtual void visit(EQExpAST& ast) = 0;
    virtual void visit(LANDExpAST& ast) = 0;
    virtual void visit(LORExpAST& ast) = 0;
};

#endif //TEST_AST_H
