#include "genIR.h"


#define CONST_INT(num) new ConstantInt(module->int32_ty_, num)
#define CONST_FLOAT(num) new ConstantFloat(module->float32_ty_, num)
#define VOID_T (module->void_ty_)
#define INT1_T (module->int1_ty_)
#define INT32_T (module->int32_ty_)
#define FLOAT_T (module->float32_ty_)
#define INT32PTR_T (module->get_pointer_type(module->int32_ty_))
#define FLOATPTR_T (module->get_pointer_type(module->float32_ty_))

Type *curType;                      
bool isConst;                      
bool useConst = false;             
std::vector<Type *> params;         
std::vector<std::string> paramNames; 
Value *retAlloca = nullptr;          
BasicBlock *retBB = nullptr;   
bool isNewFunc = false; 
bool requireLVal = false; 
Function *currentFunction = nullptr; 
Value *recentVal = nullptr;         
BasicBlock *functionBB = nullptr;   
BasicBlock *whileCondBB = nullptr;  
BasicBlock *trueBB = nullptr; 
BasicBlock *falseBB = nullptr; 
BasicBlock *whileFalseBB;   
int id = 1;                 
bool has_br = false;       
bool is_single_exp = false; 


void GenIR::checkInitType() const {
  if (curType == INT32_T) {
    if (dynamic_cast<ConstantFloat *>(recentVal)) {
      auto temp = dynamic_cast<ConstantFloat *>(recentVal);
      recentVal = CONST_INT((int)temp->value_);
    } else if (recentVal->type->tid_ == Type::FloatTyID) {
      recentVal = builder->create_fptosi(recentVal, INT32_T);
    }
  } else if (curType == FLOAT_T) {
    if (dynamic_cast<ConstantInt *>(recentVal)) {
      auto temp = dynamic_cast<ConstantInt *>(recentVal);
      recentVal = CONST_FLOAT(temp->value_);
    } else if (recentVal->type->tid_ == Type::IntegerTyID) {
      recentVal = builder->create_sitofp(recentVal, FLOAT_T);
    }
  }
}

void GenIR::visit(COMPUnitAST &ast) {
  for (const auto &def : ast.declDefList) {
    def->ACcept(*this);
  }
}

void GenIR::visit(DECLDefAST &ast) {
  if (ast.Decl != nullptr) {
    ast.Decl->ACcept(*this);
  } else {
    ast.funcDef->ACcept(*this);
  }
}

void GenIR::visit(DECLAST &ast) {
  isConst = ast.isConst;
  curType = ast.bType == TYPE_INT ? INT32_T : FLOAT_T;
  for (auto &def : ast.defList) {
    def->ACcept(*this);
  }
}

void GenIR::visit(DEFAST &ast) {
  string varName = *ast.id;
  if (scope.in_global()) {
    if (ast.arrays.empty()) {     
      if (ast.initVal == nullptr) { 
        if (isConst)
          cout << "no initVal when define const!"
               << endl; 
        GlobalVariable *var;
        if (curType == INT32_T)
          var = new GlobalVariable(varName, module.get(), curType, false,
                                   CONST_INT(0));
        else
          var = new GlobalVariable(varName, module.get(), curType, false,
                                   CONST_FLOAT(0));
        scope.push(varName, var);
      } else { 
        useConst = true;
        ast.initVal->ACcept(*this);
        useConst = false;
        checkInitType();
        if (isConst) {
          scope.push(varName, recentVal); 
        } else {                         
          auto initializer = static_cast<Constant *>(recentVal);
          GlobalVariable *var;
          var = new GlobalVariable(varName, module.get(), curType, false,
                                   initializer);
          scope.push(varName, var);
        }
      }
    } else {                  
      vector<int> dimensions; 
      useConst = true;
      for (auto &exp : ast.arrays) {
        exp->ACcept(*this);
        int dimension = dynamic_cast<ConstantInt *>(recentVal)->value_;
        dimensions.push_back(dimension);
      }
      useConst = false;
      vector<ArrayType *> arrayTys(
          dimensions.size()); 
      for (int i = dimensions.size() - 1; i >= 0; i--) {
        if (i == dimensions.size() - 1)
          arrayTys[i] = module->get_array_type(curType, dimensions[i]);
        else
          arrayTys[i] = module->get_array_type(arrayTys[i + 1], dimensions[i]);
      }
      if (ast.initVal == nullptr || ast.initVal->initValList.empty()) {
        auto init = new ConstantZero(arrayTys[0]);
        auto var = new GlobalVariable(varName, module.get(), arrayTys[0],
                                      isConst, init);
        scope.push(varName, var);
      } else {
        useConst = true; 
        auto init =
            globalInit(dimensions, arrayTys, 0, ast.initVal->initValList);
        useConst = false;
        auto var = new GlobalVariable(varName, module.get(), arrayTys[0],
                                      isConst, init);
        scope.push(varName, var);
      }
    }
    return;
  }

  if (ast.arrays.empty()) {    
    if (ast.initVal == nullptr) { 
      if (isConst)
        cout << "no initVal when define const!" << endl; 
      else {                                           
        auto _backupBB = builder->get_insert_block();
        builder->set_insert_point(functionBB);   
        AllocaInst *varAlloca;
        varAlloca = builder->create_alloca(curType);
        builder->set_insert_point(_backupBB); 
        scope.push(varName, varAlloca);
      }
    } else { 
      ast.initVal->ACcept(*this);
      checkInitType();
      if (isConst) {
        scope.push(varName, recentVal); 
      } else {
        auto _backupBB = builder->get_insert_block();
        builder->set_insert_point(functionBB); 
        AllocaInst *varAlloca;
        varAlloca = builder->create_alloca(curType);
        builder->set_insert_point(_backupBB);
        scope.push(varName, varAlloca);
        builder->create_store(recentVal, varAlloca);
      }
    }
  } else { 
    vector<int> dimensions(ast.arrays.size()),
        dimensionsCnt((ast.arrays.size())); 
    int totalByte = 1; 
    useConst = true;
    for (int i = dimensions.size() - 1; i >= 0; i--) {
      ast.arrays[i]->ACcept(*this);
      int dimension = dynamic_cast<ConstantInt *>(recentVal)->value_;
      totalByte *= dimension;
      dimensions[i] = dimension;
      dimensionsCnt[i] = totalByte;
    }
    totalByte *= 4; 
    useConst = false;
    ArrayType *arrayTy = nullptr; 
    for (int i = dimensions.size() - 1; i >= 0; i--) {
      if (i == dimensions.size() - 1)
        arrayTy = module->get_array_type(curType, dimensions[i]);
      else
        arrayTy = module->get_array_type(arrayTy, dimensions[i]);
    }
    auto arrayAlloc = builder->create_alloca(arrayTy);
    scope.push(varName, arrayAlloc);
    if (ast.initVal == nullptr) { 
      if (isConst)
        cout << "no initVal when define const!" << endl; 
      return; 
    }
    Value *i32P = builder->create_bitcast(arrayAlloc, INT32PTR_T);
    auto memclr = scope.find("memclr");
    builder->create_call(
        memclr, {i32P, CONST_INT(totalByte)});
    if (ast.initVal->initValList.empty())
      return;
    vector<Value *> idxs(dimensions.size() + 1);
    for (int i = 0; i < dimensions.size() + 1; i++) {
      idxs[i] = CONST_INT(0);
    }
    Value *ptr = builder->create_gep(arrayAlloc, idxs); 
    localInit(ptr, ast.initVal->initValList, dimensionsCnt, 1);
  }
}

int GenIR::getNextDim(vector<int> &elementsCnts, int up) {
  for (int i = elementsCnts.size() - 1; i > up; i--) {
    if (elementsCnts[i] != 0)
      return i + 1;
  }
  return up + 1;
}

void GenIR::mergeElements(vector<int> &dimensions,
                          vector<ArrayType *> &arrayTys, int up, int dimAdd,
                          vector<Constant *> &elements,
                          vector<int> &elementsCnts) {
  for (int i = dimAdd; i > up; i--) {
    if (elementsCnts[i] % dimensions[i] == 0) {
      vector<Constant *> temp;
      temp.assign(elements.end() - dimensions[i], elements.end());
      elements.erase(elements.end() - dimensions[i], elements.end());
      elements.push_back(new ConstantArray(arrayTys[i], temp));
      elementsCnts[i] = 0;
      elementsCnts[i - 1]++;
    } else
      break;
  }
}
void GenIR::finalMerge(vector<int> &dimensions, vector<ArrayType *> &arrayTys,
                       int up, vector<Constant *> &elements,
                       vector<int> &elementsCnts) const {
  for (int i = dimensions.size() - 1; i >= up; i--) {
    while (elementsCnts[i] % dimensions[i] != 0) { 
      if (i == dimensions.size() - 1) {
        if (curType == INT32_T) {
          elements.push_back(CONST_INT(0));
        } else {
          elements.push_back(CONST_FLOAT(0));
        }
      } else {
        elements.push_back(new ConstantZero(arrayTys[i + 1]));
      }
      elementsCnts[i]++;
    }
    if (elementsCnts[i] != 0) {
      vector<Constant *> temp;
      temp.assign(elements.end() - dimensions[i], elements.end());
      elements.erase(elements.end() - dimensions[i], elements.end());
      elements.push_back(new ConstantArray(arrayTys[i], temp));
      elementsCnts[i] = 0;
      if (i != up)
        elementsCnts[i - 1]++;
    }
  }
}

ConstantArray *GenIR::globalInit(vector<int> &dimensions,
                                 vector<ArrayType *> &arrayTys, int up,
                                 vector<unique_ptr<INITValAST>> &list) {
  vector<int> elementsCnts(dimensions.size());
  vector<Constant *> elements;                
  int dimAdd;
  for (auto &val : list) {
    if (val->exp != nullptr) {
      dimAdd = dimensions.size() - 1;
      val->exp->ACcept(*this);
      checkInitType();
      elements.push_back((ConstantInt *)recentVal);
    } else {
      auto nextUp = getNextDim(elementsCnts, up);
      dimAdd = nextUp - 1; 
      if (nextUp == dimensions.size())
        cout << "initial invalid" << endl;
      if (val->initValList.empty()) {
        elements.push_back(new ConstantZero(arrayTys[nextUp]));
      } else {
        auto temp = globalInit(dimensions, arrayTys, nextUp, val->initValList);
        elements.push_back(temp);
      }
    }
    elementsCnts[dimAdd]++;
    mergeElements(dimensions, arrayTys, up, dimAdd, elements, elementsCnts);
  }
  finalMerge(dimensions, arrayTys, up, elements, elementsCnts);
  return static_cast<ConstantArray *>(elements[0]);
}

int GenIR::getNextDim(vector<int> &dimensionsCnt, int up, int cnt) {
  for (int i = up; i < dimensionsCnt.size(); i++) {
    if (cnt % dimensionsCnt[i] == 0)
      return i;
  }
  return 0;
}

void GenIR::localInit(Value *ptr, vector<unique_ptr<INITValAST>> &list,
                      vector<int> &dimensionsCnt, int up) {
  int cnt = 0;
  Value *tempPtr = ptr;
  for (auto &initVal : list) {
    if (initVal->exp) {
      if (cnt == 0)
        cnt++; 
      else
        tempPtr = builder->create_gep(ptr, {CONST_INT(cnt++)});
      initVal->exp->ACcept(*this);
      checkInitType();
      builder->create_store(recentVal, tempPtr);
    } else {
      auto nextUp = getNextDim(dimensionsCnt, up, cnt);
      if (nextUp == 0)
        cout << "initial invalid!" << endl;
      if (!initVal->initValList.empty()) {
        if (cnt != 0)
          tempPtr = builder->create_gep(
              ptr, {CONST_INT(cnt)});
        localInit(tempPtr, initVal->initValList, dimensionsCnt, nextUp);
      }
      cnt += dimensionsCnt[nextUp]; 
    }
  }
}

void GenIR::visit(INITValAST &ast) {
  if (ast.exp != nullptr) {
    ast.exp->ACcept(*this);
  }
}

void GenIR::visit(FUNCDefAST &ast) {
  isNewFunc = true;
  params.clear();
  paramNames.clear();
  Type *retType;
  if (ast.funcType == TYPE_INT)
    retType = INT32_T;
  else if (ast.funcType == TYPE_FLOAT)
    retType = FLOAT_T;
  else
    retType = VOID_T;

  for (auto &funcFParam : ast.funcFParamList) {
    funcFParam->ACcept(*this);
  }
  auto funTy = new FunctionType(retType, params);
  auto func = new Function(funTy, *ast.id, module.get());
  currentFunction = func;
  scope.push(*ast.id, func); 
  scope.enter();

  std::vector<Value *> args; 
  for (auto arg = func->arguments_.begin(); arg != func->arguments_.end();
       arg++)
    args.push_back(*arg);

  auto bb = new BasicBlock(module.get(), "label_entry", func);
  builder->BB_ = bb;
  functionBB = bb;
  for (int i = 0; i < (int)(paramNames.size()); i++) {
    auto alloc = builder->create_alloca(params[i]);
    builder->create_store(args[i], alloc);          
    scope.push(paramNames[i], alloc);              
  }
  retBB = new BasicBlock(module.get(), "label_ret", func);
  if (retType == VOID_T) {
    builder->BB_ = retBB;
    builder->create_void_ret();
  } else {
    retAlloca = builder->create_alloca(retType); 
    builder->BB_ = retBB;
    auto retLoad = builder->create_load(retAlloca);
    builder->create_ret(retLoad);
  }
  builder->BB_ = bb;
  has_br = false;
  ast.block->ACcept(*this);

  if (!builder->BB_->get_terminator())
    builder->create_br(retBB);
}

void GenIR::visit(FUNCFParamAST &ast) {
  Type *paramType;
  if (ast.bType == TYPE_INT)
    paramType = INT32_T;
  else
    paramType = FLOAT_T;
  if (ast.isArray) {
    useConst = true;
    for (int i = ast.arrays.size() - 1; i >= 0; i--) {
      ast.arrays[i]->ACcept(*this);
      paramType =
          module->get_array_type(paramType, ((ConstantInt *)recentVal)->value_);
    }
    useConst = false;
    paramType = module->get_pointer_type(paramType);
  }
  params.push_back(paramType);
  paramNames.push_back(*ast.id);
}

void GenIR::visit(BLOCKAST &ast) {
  if (isNewFunc)
    isNewFunc = false;
  else {
    scope.enter();
  }
  for (auto &item : ast.blockItemList) {
    if (has_br)
      break; 
    item->ACcept(*this);
  }

  scope.exit();
}

void GenIR::visit(BLOCKItemAST &ast) {
  if (ast.decl != nullptr) {
    ast.decl->ACcept(*this);
  } else {
    ast.stmt->ACcept(*this);
  }
}

void GenIR::visit(STMTAST &ast) {
  switch (ast.sType) {
  case SEMI:
    break;
  case ASS: {
    requireLVal = true;
    ast.lVal->ACcept(*this);
    auto lVal = recentVal;
    auto lValType = static_cast<PointerType *>(lVal->type)->contained_;
    ast.exp->ACcept(*this);
    auto rVal = recentVal;
    if (lValType != recentVal->type) {
      if (lValType == FLOAT_T) {
        rVal = builder->create_sitofp(recentVal, FLOAT_T);
      } else {
        rVal = builder->create_fptosi(recentVal, INT32_T);
      }
    }
    builder->create_store(rVal, lVal);
    break;
  }
  case EXP:
    is_single_exp = true;
    ast.exp->ACcept(*this);
    is_single_exp = false;
    break;
  case CONT:
    builder->create_br(whileCondBB);
    has_br = true;
    break;
  case BRE:
    builder->create_br(whileFalseBB);
    has_br = true;
    break;
  case RET:
    ast.returnStmt->ACcept(*this);
    break;
  case BLK:
    ast.block->ACcept(*this);
    break;
  case SEL:
    ast.selectStmt->ACcept(*this);
    break;
  case ITER:
    ast.iterationStmt->ACcept(*this);
    break;
  }
}

void GenIR::visit(RETURNStmtAST &ast) {
  if (ast.exp == nullptr) {
    recentVal = builder->create_br(retBB);
  } else {
    ast.exp->ACcept(*this);
    if (recentVal->type == FLOAT_T &&
        currentFunction->get_return_type() == INT32_T) {
      auto temp = builder->create_fptosi(recentVal, INT32_T);
      builder->create_store(temp, retAlloca);
    } else if (recentVal->type == INT32_T &&
               currentFunction->get_return_type() == FLOAT_T) {
      auto temp = builder->create_sitofp(recentVal, FLOAT_T);
      builder->create_store(temp, retAlloca);
    } else
      builder->create_store(recentVal, retAlloca);
    recentVal = builder->create_br(retBB);
  }
  has_br = true;
}

void GenIR::visit(SELECTStmtAST &ast) {
  auto tempTrue = trueBB;
  auto tempFalse = falseBB;

  trueBB = new BasicBlock(module.get(), to_string(id++), currentFunction);
  falseBB = new BasicBlock(module.get(), to_string(id++), currentFunction);
  BasicBlock *nextIf;
  if (ast.elseStmt == nullptr)
    nextIf = falseBB;
  else
    nextIf = new BasicBlock(module.get(), to_string(id++), currentFunction);
  ast.cond->ACcept(*this);
  if (recentVal->type == INT32_T) {
    recentVal = builder->create_icmp(ICmpInst::ICMP_NE, recentVal, CONST_INT(0));
  } else if (recentVal->type == FLOAT_T) {
    recentVal = builder->create_fcmp(FCmpInst::FCMP_UNE,recentVal, CONST_FLOAT(0));
  }
  builder->create_cond_br(recentVal, trueBB, falseBB);

  builder->BB_ = trueBB;
  has_br = false;
  ast.ifStmt->ACcept(*this);
  if (!builder->BB_->get_terminator()) {
    builder->create_br(nextIf);
  }

  if (ast.elseStmt != nullptr) { 
    builder->BB_ = falseBB;
    has_br = false;
    ast.elseStmt->ACcept(*this);
    if (!builder->BB_->get_terminator()) {
      builder->create_br(nextIf);
    }
  }

  builder->BB_ = nextIf;
  has_br = false;
  trueBB = tempTrue;
  falseBB = tempFalse;
}

void GenIR::visit(ITERATIONStmtAST &ast) {
  auto tempTrue = trueBB;
  auto tempFalse = falseBB; 
  auto tempCond = whileCondBB;
  auto tempWhileFalseBB =
      whileFalseBB; 

  whileCondBB = new BasicBlock(module.get(), to_string(id++), currentFunction);
  trueBB = new BasicBlock(module.get(), to_string(id++), currentFunction);
  falseBB = new BasicBlock(module.get(), to_string(id++), currentFunction);
  whileFalseBB = falseBB;

  builder->create_br(whileCondBB);
  builder->BB_ = whileCondBB; 
  has_br = false;
  ast.cond->ACcept(*this);
  if (recentVal->type == INT32_T) {
    recentVal = builder->create_icmp(ICmpInst::ICMP_NE, recentVal, CONST_INT(0));
  } else if (recentVal->type == FLOAT_T) {
    recentVal = builder->create_fcmp(FCmpInst::FCMP_UNE, recentVal, CONST_FLOAT(0.0));
  }
  builder->create_cond_br(recentVal, trueBB, falseBB);

  builder->BB_ = trueBB;
  has_br = false;
  ast.stmt->ACcept(*this);
  if (!builder->BB_->get_terminator()) {
    builder->create_br(whileCondBB);
  }

  builder->BB_ = falseBB;
  has_br = false;

  trueBB = tempTrue;
  falseBB = tempFalse;
  whileCondBB = tempCond;
  whileFalseBB = tempWhileFalseBB;
}

bool GenIR::checkCalType(Value *val[], int intVal[], float floatVal[]) {
  bool resultIsInt = false;
  if (dynamic_cast<ConstantInt *>(val[0]) &&
      dynamic_cast<ConstantInt *>(val[1])) {
    resultIsInt = true;
    intVal[0] = dynamic_cast<ConstantInt *>(val[0])->value_;
    intVal[1] = dynamic_cast<ConstantInt *>(val[1])->value_;
  } else { 
    if (dynamic_cast<ConstantInt *>(val[0]))
      floatVal[0] = dynamic_cast<ConstantInt *>(val[0])->value_;
    else
      floatVal[0] = dynamic_cast<ConstantFloat *>(val[0])->value_;
    if (dynamic_cast<ConstantInt *>(val[1]))
      floatVal[1] = dynamic_cast<ConstantInt *>(val[1])->value_;
    else
      floatVal[1] = dynamic_cast<ConstantFloat *>(val[1])->value_;
  }
  return resultIsInt;
}

void GenIR::checkCalType(Value *val[]) {
  if (val[0]->type == INT1_T) {
    val[0] = builder->create_zext(val[0], INT32_T);
  }
  if (val[1]->type == INT1_T) {
    val[1] = builder->create_zext(val[1], INT32_T);
  }
  if (val[0]->type == INT32_T && val[1]->type == FLOAT_T) {
    val[0] = builder->create_sitofp(val[0], FLOAT_T);
  }
  if (val[1]->type == INT32_T && val[0]->type == FLOAT_T) {
    val[1] = builder->create_sitofp(val[1], FLOAT_T);
  }
}

void GenIR::visit(ADDExpAST &ast) {
  if (ast.addExp == nullptr) {
    ast.mulExp->ACcept(*this);
    return;
  }

  Value *val[2]; 
  ast.addExp->ACcept(*this);
  val[0] = recentVal;
  ast.mulExp->ACcept(*this);
  val[1] = recentVal;

  if (useConst) {
    int intVal[3];    
    float floatVal[3]; 
    bool resultIsInt = checkCalType(val, intVal, floatVal);
    switch (ast.op) {
    case AOP_ADD:
      intVal[2] = intVal[0] + intVal[1];
      floatVal[2] = floatVal[0] + floatVal[1];
      break;
    case AOP_MINUS:
      intVal[2] = intVal[0] - intVal[1];
      floatVal[2] = floatVal[0] - floatVal[1];
      break;
    }
    if (resultIsInt)
      recentVal = CONST_INT(intVal[2]);
    else
      recentVal = CONST_FLOAT(floatVal[2]);
    return;
  }
  checkCalType(val);
  if (val[0]->type == INT32_T) {
    switch (ast.op) {
    case AOP_ADD:
      recentVal = builder->create_binary(Instruction::Add, val[0], val[1]);
      break;
    case AOP_MINUS:
      recentVal = builder->create_binary(Instruction::Sub, val[0], val[1]);
      break;
    }
  } else {
    switch (ast.op) {
    case AOP_ADD:
      recentVal = builder->create_fbinary(Instruction::FAdd, val[0], val[1]);
      break;
    case AOP_MINUS:
      recentVal = builder->create_fbinary(Instruction::FSub, val[0], val[1]);
      break;
    }
  }
}

void GenIR::visit(MULExpAST &ast) {
  if (ast.mulExp == nullptr) {
    ast.unaryExp->ACcept(*this);
    return;
  }

  Value *val[2]; 
  ast.mulExp->ACcept(*this);
  val[0] = recentVal;
  ast.unaryExp->ACcept(*this);
  val[1] = recentVal;
  if (useConst) {
    int intVal[3];     
    float floatVal[3]; 
    bool resultIsInt = checkCalType(val, intVal, floatVal);
    switch (ast.op) {
    case MOP_MUL:
      intVal[2] = intVal[0] * intVal[1];
      floatVal[2] = floatVal[0] * floatVal[1];
      break;
    case MOP_DIV:
      intVal[2] = intVal[0] / intVal[1];
      floatVal[2] = floatVal[0] / floatVal[1];
      break;
    case MOP_MOD:
      intVal[2] = intVal[0] % intVal[1];
      break;
    }
    if (resultIsInt)
      recentVal = CONST_INT(intVal[2]);
    else
      recentVal = CONST_FLOAT(floatVal[2]);
    return;
  }

  checkCalType(val);
  if (val[0]->type == INT32_T) {
    switch (ast.op) {
    case MOP_MUL:
      recentVal = builder->create_binary(Instruction::Mul, val[0], val[1]);
      break;
    case MOP_DIV:
      recentVal = builder->create_binary(Instruction::SDiv, val[0], val[1]);
      break;
    case MOP_MOD:
      recentVal = builder->create_binary(Instruction::SRem, val[0], val[1]);
      break;
    }
  } else {
    switch (ast.op) {
    case MOP_MUL:
      recentVal = builder->create_fbinary(Instruction::FMul, val[0], val[1]);
      break;
    case MOP_DIV:
      recentVal = builder->create_fbinary(Instruction::FDiv, val[0], val[1]);
      break;
    case MOP_MOD: 
      break;
    }
  }
}

void GenIR::visit(UNARYExpAST &ast) {
  if (useConst) {
    if (ast.primaryExp) {
      ast.primaryExp->ACcept(*this);
    } else if (ast.unaryExp) {
      ast.unaryExp->ACcept(*this);
      if (ast.op == UOP_MINUS) {
        if (dynamic_cast<ConstantInt *>(recentVal)) {
          auto temp = (ConstantInt *)recentVal;
          temp->value_ = -temp->value_;
          recentVal = temp;
        } else {
          auto temp = (ConstantFloat *)recentVal;
          temp->value_ = -temp->value_;
          recentVal = temp;
        }
      }
    } else {
      cout << "Function call in ConstExp!" << endl;
    }
    return;
  }
  if (ast.primaryExp != nullptr) {
    ast.primaryExp->ACcept(*this);
  } else if (ast.call != nullptr) {
    ast.call->ACcept(*this);
  } else {
    ast.unaryExp->ACcept(*this);
    if (recentVal->type == VOID_T)
      return;
    else if (recentVal->type == INT1_T) 
      recentVal = builder->create_zext(recentVal, INT32_T);

    if (recentVal->type == INT32_T) {
      switch (ast.op) {
      case UOP_MINUS:
        recentVal = builder->create_binary(Instruction::Sub, CONST_INT(0), recentVal);
        break;
      case UOP_NOT:
        recentVal = builder->create_icmp(ICmpInst::ICMP_EQ, recentVal, CONST_INT(0));
        break;
      case UOP_ADD:
        break;
      }
    } else {
      switch (ast.op) {
      case UOP_MINUS:
        recentVal = builder->create_fbinary(Instruction::FSub, CONST_FLOAT(0), recentVal);
        break;
      case UOP_NOT:
        recentVal = builder->create_fcmp(FCmpInst::FCMP_UEQ, recentVal, CONST_FLOAT(0));
        break;
      case UOP_ADD:
        break;
      }
    }
  }
}

void GenIR::visit(PRIMARYExpAST &ast) {
  if (ast.exp) {
    ast.exp->ACcept(*this);
  } else if (ast.lval) {
    ast.lval->ACcept(*this);
  } else if (ast.number) {
    ast.number->ACcept(*this);
  }
}

void GenIR::visit(LValAST &ast) {
  bool isTrueLVal = requireLVal; 
  requireLVal = false;
  auto var = scope.find(*ast.id);
  if (scope.in_global()) {
    if (ast.arrays.empty()) {
      recentVal = var;
      return;
    }
    vector<int> index;
    for (auto &exp : ast.arrays) {
      exp->ACcept(*this);
      index.push_back(dynamic_cast<ConstantInt *>(recentVal)->value_);
    }
    recentVal =
        ((GlobalVariable *)var)->init_val_; 
    for (int i : index) {
      if (dynamic_cast<ConstantZero *>(recentVal)) {
        Type *arrayTy = recentVal->type;
        while (dynamic_cast<ArrayType *>(arrayTy)) {
          arrayTy = dynamic_cast<ArrayType *>(arrayTy)->contained_;
        }
        if (arrayTy == INT32_T)
          recentVal = CONST_INT(0);
        else
          recentVal = CONST_FLOAT(0);
        return;
      }
      if (dynamic_cast<ConstantArray *>(recentVal)) {
        recentVal = ((ConstantArray *)recentVal)->const_array[i];
      }
    }
    return;
  }

  if (var->type->tid_ == Type::IntegerTyID ||
      var->type->tid_ == Type::FloatTyID) { 
    recentVal = var;
    return;
  }
  Type *varType =
      static_cast<PointerType *>(var->type)->contained_; 
  if (!ast.arrays.empty()) {                              
    vector<Value *> idxs;
    for (auto &exp : ast.arrays) {
      exp->ACcept(*this);
      idxs.push_back(recentVal);
    }
    if (varType->tid_ == Type::PointerTyID) {
      var = builder->create_load(var);
    } else if (varType->tid_ == Type::ArrayTyID) {
      idxs.insert(idxs.begin(), CONST_INT(0));
    }
    var = builder->create_gep(var, idxs); 
    varType = ((PointerType *)var->type)->contained_;
  }
  if (varType->tid_ == Type::ArrayTyID) {
    recentVal = builder->create_gep(var, {CONST_INT(0), CONST_INT(0)});
  } else if (!isTrueLVal) { 
    recentVal = builder->create_load(var);
  } else {
    recentVal = var;
  }
}

void GenIR::visit(NUMBERAST &ast) {
  if (ast.isInt)
    recentVal = CONST_INT(ast.intval);
  else
    recentVal = CONST_FLOAT(ast.floatval);
}

void GenIR::visit(CALLAST &ast) {
  auto fun = (Function *)scope.find(*ast.id);
  if (fun->basic_blocks_.size() && !is_single_exp)
    fun->use_ret_cnt++;
  is_single_exp = false;
  vector<Value *> args;
  for (int i = 0; i < ast.funcCParamList.size(); i++) {
    ast.funcCParamList[i]->ACcept(*this);
    if (recentVal->type == INT32_T && fun->arguments_[i]->type == FLOAT_T) {
      recentVal = builder->create_sitofp(recentVal, FLOAT_T);
    } else if (recentVal->type == FLOAT_T &&
               fun->arguments_[i]->type == INT32_T) {
      recentVal = builder->create_fptosi(recentVal, INT32_T);
    }
    args.push_back(recentVal);
  }
  recentVal = builder->create_call(fun, args);
}

void GenIR::visit(RELExpAST &ast) {
  if (ast.relExp == nullptr) {
    ast.addExp->ACcept(*this);
    return;
  }
  Value *val[2];
  ast.relExp->ACcept(*this);
  val[0] = recentVal;
  ast.addExp->ACcept(*this);
  val[1] = recentVal;
  checkCalType(val);
  if (val[0]->type == INT32_T) {
    switch (ast.op) {
    case ROP_LTE:
      recentVal = builder->create_icmp(ICmpInst::ICMP_SLE, val[0], val[1]);
      break;
    case ROP_LT:
      recentVal = builder->create_icmp(ICmpInst::ICMP_SLT, val[0], val[1]);
      break;
    case ROP_GT:
      recentVal = builder->create_icmp(ICmpInst::ICMP_SGT, val[0], val[1]);
      break;
    case ROP_GTE:
      recentVal = builder->create_icmp(ICmpInst::ICMP_SGE, val[0], val[1]);
      break;
    }
  } else {
    switch (ast.op) {
    case ROP_LTE:
      recentVal = builder->create_fcmp(FCmpInst::FCMP_ULE, val[0], val[1]);
      break;
    case ROP_LT:
      recentVal = builder->create_fcmp(FCmpInst::FCMP_ULT, val[0], val[1]);
      break;
    case ROP_GT:
      recentVal = builder->create_fcmp(FCmpInst::FCMP_UGT, val[0], val[1]);
      break;
    case ROP_GTE:
      recentVal = builder->create_fcmp(FCmpInst::FCMP_UGE, val[0], val[1]);
      break;
    }
  }
}

void GenIR::visit(EQExpAST &ast) {
  if (ast.eqExp == nullptr) {
    ast.relExp->ACcept(*this);
    return;
  }
  Value *val[2];
  ast.eqExp->ACcept(*this);
  val[0] = recentVal;
  ast.relExp->ACcept(*this);
  val[1] = recentVal;
  checkCalType(val);
  if (val[0]->type == INT32_T) {
    switch (ast.op) {
    case EOP_EQ:
      recentVal = builder->create_icmp(ICmpInst::ICMP_EQ, val[0], val[1]);
      break;
    case EOP_NEQ:
      recentVal = builder->create_icmp(ICmpInst::ICMP_NE, val[0], val[1]);
      break;
    }
  } else {
    switch (ast.op) {
    case EOP_EQ:
      recentVal = builder->create_fcmp(FCmpInst::FCMP_UEQ, val[0], val[1]);
      break;
    case EOP_NEQ:
      recentVal = builder->create_fcmp(FCmpInst::FCMP_UNE, val[0], val[1]);
      break;
    }
  }
}

void GenIR::visit(LANDExpAST &ast) {
  if (ast.lAndExp == nullptr) {
    ast.eqExp->ACcept(*this);
    return;
  }
  auto tempTrue = trueBB; 
  trueBB = new BasicBlock(module.get(), to_string(id++), currentFunction);
  ast.lAndExp->ACcept(*this);

  if (recentVal->type == INT32_T) {
    recentVal = builder->create_icmp(ICmpInst::ICMP_NE, recentVal, CONST_INT(0));
  } else if (recentVal->type == FLOAT_T) {
    recentVal = builder->create_fcmp(FCmpInst::FCMP_UNE, recentVal, CONST_FLOAT(0));
  }
  builder->create_cond_br(recentVal, trueBB, falseBB);
  builder->BB_ = trueBB;
  has_br = false;
  trueBB = tempTrue; 

  ast.eqExp->ACcept(*this);
}

void GenIR::visit(LORExpAST &ast) {
  if (ast.lOrExp == nullptr) {
    ast.lAndExp->ACcept(*this);
    return;
  }
  auto tempFalse = falseBB; 
  falseBB = new BasicBlock(module.get(), to_string(id++), currentFunction);
  ast.lOrExp->ACcept(*this);
  if (recentVal->type == INT32_T) {
    recentVal = builder->create_icmp(ICmpInst::ICMP_NE, recentVal, CONST_INT(0));
  } else if (recentVal->type == FLOAT_T) {
    recentVal = builder->create_fcmp(FCmpInst::FCMP_UNE, recentVal, CONST_FLOAT(0));
  }
  builder->create_cond_br(recentVal, trueBB, falseBB);
  builder->BB_ = falseBB;
  has_br = false;
  falseBB = tempFalse;

  ast.lAndExp->ACcept(*this);
}
