//#pragma once
#ifndef _LLVM_H
#define _LLVM_H
#include <algorithm>
#include <cassert>
#include <cstdint>
#include <iostream>
#include <list>
#include <map>
#include <set>
#include <sstream>
#include <string>
#include <utility>
#include <vector>
#include "value.h"


class Constant : public Value {
public:
  Constant(Type *ty, const std::string &name = "") : Value(ty, name) {}
  ~Constant() = default;
};


class ConstantInt : public Constant {
public:
  ConstantInt(Type *ty, int val) : Constant(ty, ""), value_(val) {}
  virtual std::string print() override;
  int value_;
};


class ConstantFloat : public Constant {
public:
  ConstantFloat(Type *ty, float val) : Constant(ty, ""), value_(val) {}
  virtual std::string print() override;
  float value_;
  std::string print32();
};

class ConstantArray : public Constant {
public:
  ConstantArray(ArrayType *ty, const std::vector<Constant *> &val)
      : Constant(ty, "") {
    this->const_array.assign(val.begin(), val.end());
  }
  ~ConstantArray() = default;
  virtual std::string print() override;
  std::vector<Constant *> const_array;
};


class ConstantZero : public Constant {
public:
  ConstantZero(Type *ty) : Constant(ty, "") {}
  virtual std::string print() override;
};

class GlobalVariable;
class Function;
class BasicBlock;
class Argument;
class Instruction;


class Module {
public:
  Function *getMainFunc();
  std::vector<GlobalVariable *> global_list_;
  std::vector<Function *> function_list_;

  IntegerType *int1_ty_;
  IntegerType *int32_ty_;
  Type *float32_ty_;
  Type *label_ty_;
  Type *void_ty_;
  std::map<Type *, PointerType *> pointer_map_;
  std::map<std::pair<Type *, int>, ArrayType *> array_map_;

  explicit Module() {
    void_ty_ = new Type(Type::VoidTyID);
    label_ty_ = new Type(Type::LabelTyID);
    int1_ty_ = new IntegerType(1);
    int32_ty_ = new IntegerType(32);
    float32_ty_ = new Type(Type::FloatTyID);
  }
  ~Module() {
    delete void_ty_;
    delete label_ty_;
    delete int1_ty_;
    delete int32_ty_;
    delete float32_ty_;
  }
  virtual std::string print();
  void add_global_variable(GlobalVariable *g) { global_list_.push_back(g); }
  void add_function(Function *f) { function_list_.push_back(f); }
  PointerType *get_pointer_type(Type *contained) {
    if (!pointer_map_.count(contained)) {
      pointer_map_[contained] = new PointerType(contained);
    }
    return pointer_map_[contained];
  }
  ArrayType *get_array_type(Type *contained, unsigned num_elements) {
    if (!array_map_.count({contained, num_elements})) {
      array_map_[{contained, num_elements}] =
          new ArrayType(contained, num_elements);
    }
    return array_map_[{contained, num_elements}];
  }

 };


class GlobalVariable : public Value {
public:
  GlobalVariable(std::string name, Module *m, Type *ty, bool is_const,
                 Constant *init = nullptr)
      : Value(m->get_pointer_type(ty), name), is_const_(is_const),
        init_val_(init) {
    m->add_global_variable(this);
  }
  virtual std::string print() override;
  bool is_const_;
  Constant *init_val_;
};


class Argument : public Value {
public:
  explicit Argument(Type *ty, const std::string &name = "",
                    Function *f = nullptr, unsigned arg_no = 0)
      : Value(ty, name), parent_(f), arg_no_(arg_no) {}
  ~Argument() {}
  virtual std::string print() override;
  Function *parent_;
  unsigned arg_no_; 
};

class Function : public Value {
public:
  std::vector<BasicBlock *> basic_blocks_; 
  std::vector<Argument *> arguments_;      
  Module *parent_;
  unsigned seq_cnt_;
  std::vector<std::set<Value *>> vreg_set_;
  int use_ret_cnt;
  
  Function(FunctionType *ty, const std::string &name, Module *parent)
      : Value(ty, name), parent_(parent), seq_cnt_(0) {
    parent->add_function(this);
    size_t num_args = ty->args_.size();
    use_ret_cnt = 0;
    for (size_t i = 0; i < num_args; i++) {
      arguments_.push_back(new Argument(ty->args_[i], "", this, i));
    }
  }
  ~Function();
  virtual std::string print() override;
  void add_basic_block(BasicBlock *bb) { basic_blocks_.push_back(bb); }
  Type *get_return_type() const {
    return static_cast<FunctionType *>(type)->result_;
  }
  bool is_declaration() { return basic_blocks_.empty(); }
  void set_instr_name();
  void remove_bb(BasicBlock *bb);
  BasicBlock *getRetBB();
};

class BasicBlock : public Value {
public:

  std::list<Instruction *> instr_list_;
  Function *parent_;
  std::vector<BasicBlock *> pre_bbs_;
  std::vector<BasicBlock *> succ_bbs_;
  std::set<BasicBlock *> dom_frontier_;
  std::set<BasicBlock *> rdom_frontier_;
  std::set<BasicBlock *> rdoms_;
  BasicBlock *idom_;
  //std::set<Value *> livein;
  //std::set<Value *> liveout;

  explicit BasicBlock(Module *m, const std::string &name, Function *parent)
      : Value(m->label_ty_, name), parent_(parent) {
    parent_->add_basic_block(this);
  }
  bool add_instruction(Instruction *instr); 
  bool add_instruction_front(Instruction *instr); 
  bool add_instruction_before_terminator(
      Instruction *instr); 
  bool add_instruction_before_inst(Instruction *new_inst,Instruction *inst);
  void add_pre_basic_block(BasicBlock *bb) { pre_bbs_.push_back(bb); }
  void add_succ_basic_block(BasicBlock *bb) { succ_bbs_.push_back(bb); }
  void remove_pre_basic_block(BasicBlock *bb) {
    pre_bbs_.erase(std::remove(pre_bbs_.begin(), pre_bbs_.end(), bb),pre_bbs_.end());
  }
  void remove_succ_basic_block(BasicBlock *bb) {
    succ_bbs_.erase(std::remove(succ_bbs_.begin(), succ_bbs_.end(), bb),
                    succ_bbs_.end());
  }
  int isDominate(
      BasicBlock
          *bb2) {
    if (!bb2 || this->parent_ != bb2->parent_)
      return -1;
    while (bb2->name_ != "label_entry") {
      if (bb2->idom_ == this)
        return 1;
      bb2 = bb2->idom_;
    }
    return 0;
  }
  
  Instruction *get_terminator();
  bool delete_instr(Instruction *instr); 
  bool remove_instr(Instruction *instr); 
  virtual std::string print() override;

};

class Instruction : public Value {
public:
  enum OpID {
     Ret = 11, Br, FNeg, Add, Sub, Mul, SDiv, SRem, UDiv, URem, FAdd, FSub, FMul, FDiv, Shl, LShr, AShr, And, Or, Xor, 
        Alloca, Load, Store, GetElementPtr, ZExt, FPtoSI, SItoFP, BitCast, ICmp, FCmp, PHI, Call
  };
 
  Instruction(Type *ty, OpID id, unsigned num_ops, BasicBlock *parent,bool before = false)
      : Value(ty, ""), op_id_(id), num_ops_(num_ops), parent_(parent) {
    operands_.resize(num_ops_,nullptr); 
    use_pos_.resize(num_ops_);
    if (!before)
      parent_->add_instruction(this);
    else
      parent_->add_instruction_front(this);
  }
  Instruction(Type *ty, OpID id, unsigned num_ops): Value(ty, ""), op_id_(id), num_ops_(num_ops), parent_(nullptr) {
    operands_.resize(num_ops_, nullptr);
    use_pos_.resize(num_ops_);
  }
  Value *get_operand(unsigned i) const { return operands_[i]; }

  void set_operand(unsigned i, Value *v) {
    operands_[i] = v;
    use_pos_[i] = v->add_use(this, i);
  }
  void add_operand(Value *v) { 
    operands_.push_back(v);
    use_pos_.emplace_back(v->add_use(this, num_ops_));
    num_ops_++;
  }
  void remove_use_of_ops() { 
    for (int i = 0; i < operands_.size(); i++) {
      operands_[i]->remove_use(use_pos_[i]);
    }
  }
  void remove_operands(int index1, int index2) {
    for (int i = index1; i <= index2; i++) {
      operands_[i]->remove_use(use_pos_[i]);
    }
    for (int i = index2 + 1; i < operands_.size(); i++) {
      for (auto &use : operands_[i]->use_list_) {
        if (use.val_ == this) {
          use.arg_no_ -= index2 - index1 + 1;
          break;
        }
      }
    }
    operands_.erase(operands_.begin() + index1, operands_.begin() + index2 + 1);
    use_pos_.erase(use_pos_.begin() + index1, use_pos_.begin() + index2 + 1);
    num_ops_ = operands_.size();
  }
  virtual std::string print() = 0;
  BasicBlock *parent_;
  OpID op_id_;
  unsigned num_ops_;
  std::vector<Value *> operands_;
  std::vector<std::list<Use>::iterator>use_pos_;
  std::vector<std::list<Instruction *>::iterator>pos_in_bb;
};


class UnaryInst : public Instruction {
public:
  UnaryInst(Type *ty, OpID op, Value *val, BasicBlock *bb)
      : Instruction(ty, op, 1, bb) {
    set_operand(0, val);
  }
  virtual std::string print() override;
};

class BinaryInst : public Instruction {
public:
  BinaryInst(Type *ty, OpID op, Value *v1, Value *v2, BasicBlock *bb): Instruction(ty, op, 2, bb) {
    set_operand(0, v1);
    set_operand(1, v2);
  }
  BinaryInst(Type *ty, OpID op, Value *v1, Value *v2, BasicBlock *bb, bool flag): Instruction(ty, op, 2) {
    set_operand(0, v1);
    set_operand(1, v2);
    this->parent_ = bb;
  }
  virtual std::string print() override;
};


class ICmpInst : public Instruction {
public:
  enum ICmpOp {
    ICMP_EQ = 32,  ICMP_NE = 33,   ICMP_UGT = 34,  ICMP_UGE = 35,  ICMP_ULT = 36,  ICMP_ULE = 37,  ICMP_SGT = 38,  ICMP_SGE = 39,  ICMP_SLT = 40,  ICMP_SLE = 41 
  };
  ICmpOp icmp_op_;
  static const std::map<ICmpInst::ICmpOp, std::string> ICmpOpName;
  ICmpInst(ICmpOp op, Value *v1, Value *v2, BasicBlock *bb)
      : Instruction(bb->parent_->parent_->int1_ty_, Instruction::ICmp, 2, bb),
        icmp_op_(op) {
    set_operand(0, v1);
    set_operand(1, v2);
  }
  virtual std::string print() override;
  
};

class FCmpInst : public Instruction {
public:
  enum FCmpOp {
      FCMP_FALSE = 10,  FCMP_OEQ = 11,    FCMP_OGT = 12,    FCMP_OGE = 13,    FCMP_OLT = 14,    FCMP_OLE = 15,   FCMP_ONE = 16,    FCMP_ORD = 17,    FCMP_UNO = 18,    
        FCMP_UEQ = 19,   FCMP_UGT = 20,    FCMP_UGE = 21,    FCMP_ULT = 22,    FCMP_ULE = 23,    FCMP_UNE = 24,    FCMP_TRUE = 25    
  };
  static const std::map<FCmpInst::FCmpOp, std::string> FCmpOpName;
  FCmpOp fcmp_op_;
  FCmpInst(FCmpOp op, Value *v1, Value *v2, BasicBlock *bb)
      : Instruction(bb->parent_->parent_->int1_ty_, Instruction::FCmp, 2, bb),
        fcmp_op_(op) {
    set_operand(0, v1);
    set_operand(1, v2);
  }
  virtual std::string print() override;
 
};

class CallInst : public Instruction {
public:
  CallInst(Function *func, std::vector<Value *> args, BasicBlock *bb)
      : Instruction(static_cast<FunctionType *>(func->type)->result_,
                    Instruction::Call, args.size() + 1, bb) {
    int num_ops = args.size() + 1;
    for (int i = 0; i < num_ops - 1; i++) {
      set_operand(i, args[i]);
    }
    set_operand(num_ops - 1, func);
  }
  virtual std::string print() override;
};

class BranchInst : public Instruction {
public:
  BranchInst(Value *cond, BasicBlock *if_true, BasicBlock *if_false,BasicBlock *bb)
      : Instruction(if_true->parent_->parent_->void_ty_, Instruction::Br, 3,bb) {
    if_true->add_pre_basic_block(bb);
    if_false->add_pre_basic_block(bb);
    bb->add_succ_basic_block(if_false);
    bb->add_succ_basic_block(if_true);
    set_operand(0, cond);
    set_operand(1, if_true);
    set_operand(2, if_false);
  }
  BranchInst(BasicBlock *if_true, BasicBlock *bb): Instruction(if_true->parent_->parent_->void_ty_, Instruction::Br, 1,bb) {
    if_true->add_pre_basic_block(bb);
    bb->add_succ_basic_block(if_true);
    set_operand(0, if_true);
  }
  virtual std::string print() override;
};

class ReturnInst : public Instruction {
public:
  ReturnInst(Value *val, BasicBlock *bb)
      : Instruction(bb->parent_->parent_->void_ty_, Instruction::Ret, 1, bb) {
    set_operand(0, val);
  }
  ReturnInst(Value *val, BasicBlock *bb, bool flag)
      : Instruction(bb->parent_->parent_->void_ty_, Instruction::Ret, 1) {
    set_operand(0, val);
    this->parent_ = bb;
  }
  ReturnInst(BasicBlock *bb)
      : Instruction(bb->parent_->parent_->void_ty_, Instruction::Ret, 0, bb) {}
  virtual std::string print() override;
};

class GetElementPtrInst : public Instruction {
public:
  GetElementPtrInst(Value *ptr, std::vector<Value *> idxs, BasicBlock *bb)
      : Instruction(bb->parent_->parent_->get_pointer_type(
                        get_GEP_return_type(ptr, idxs.size())),
                    Instruction::GetElementPtr, idxs.size() + 1, bb) {
    set_operand(0, ptr);
    for (size_t i = 0; i < idxs.size(); i++) {
      set_operand(i + 1, idxs[i]);
    }
  }
  Type *get_GEP_return_type(Value *ptr, size_t idxs_size) {
    Type *ty =
        static_cast<PointerType *>(ptr->type)->contained_;
    if (ty->tid_ == Type::ArrayTyID) {
      ArrayType *arr_ty = static_cast<ArrayType *>(ty);
      for (size_t i = 1; i < idxs_size; i++) {
        ty = arr_ty->contained_; 
        if (ty->tid_ == Type::ArrayTyID) {
          arr_ty = static_cast<ArrayType *>(ty);
        }
      }
    }
    return ty;
  }
  virtual std::string print() override;
};




class LoadInst : public Instruction {
public:
  LoadInst(Value *ptr, BasicBlock *bb)
      : Instruction(static_cast<PointerType *>(ptr->type)->contained_,
                    Instruction::Load, 1, bb) {
    set_operand(0, ptr);
  }
  virtual std::string print() override;
};

class StoreInst : public Instruction {
public:
  StoreInst(Value *val, Value *ptr, BasicBlock *bb)
      : Instruction(bb->parent_->parent_->void_ty_, Instruction::Store, 2, bb) {
    set_operand(0, val);
    set_operand(1, ptr);
  }

  StoreInst(Value *val, Value *ptr, BasicBlock *bb, bool)
      : Instruction(bb->parent_->parent_->void_ty_, Instruction::Store, 2) {
    set_operand(0, val);
    set_operand(1, ptr);
    this->parent_ = bb;
  }

  virtual std::string print() override;
};

class AllocaInst : public Instruction {
public:
  Type *alloca_ty_;
  AllocaInst(Type *ty, BasicBlock *bb)
      : Instruction(bb->parent_->parent_->get_pointer_type(ty),
                    Instruction::Alloca, 0, bb, true),
        alloca_ty_(ty) {}

  AllocaInst(Type *ty, BasicBlock *bb, bool)
      : Instruction(bb->parent_->parent_->get_pointer_type(ty),
                    Instruction::Alloca, 0),
        alloca_ty_(ty) {
    this->parent_ = bb;
  }

  virtual std::string print() override;
  
};

class ZextInst : public Instruction {
public:
  Type *dest_ty_;
  ZextInst(OpID op, Value *val, Type *ty, BasicBlock *bb)
      : Instruction(ty, op, 1, bb), dest_ty_(ty) {
    set_operand(0, val);
  }
  virtual std::string print() override;
  
};

class FpToSiInst : public Instruction {
public:
  Type *dest_ty_;
  FpToSiInst(OpID op, Value *val, Type *ty, BasicBlock *bb)
      : Instruction(ty, op, 1, bb), dest_ty_(ty) {
    set_operand(0, val);
  }
  virtual std::string print() override;
  
};

class SiToFpInst : public Instruction {
public:
  Type *dest_ty_;
  SiToFpInst(OpID op, Value *val, Type *ty, BasicBlock *bb): Instruction(ty, op, 1, bb), dest_ty_(ty) {
    set_operand(0, val);
  }
  virtual std::string print() override;
  
};



class PhiInst : public Instruction {
public:
  Value *l_val_;
  PhiInst(OpID op, std::vector<Value *> vals, std::vector<BasicBlock *> val_bbs,Type *ty, BasicBlock *bb): Instruction(ty, op, 2 * vals.size()) {
    for (int i = 0; i < vals.size(); i++) {
      set_operand(2 * i, vals[i]);
      set_operand(2 * i + 1, val_bbs[i]);
    }
    this->parent_ = bb;
  }
  static PhiInst *create_phi(Type *ty, BasicBlock *bb) {
    std::vector<Value *> vals;
    std::vector<BasicBlock *> val_bbs;
    return new PhiInst(Instruction::PHI, vals, val_bbs, ty, bb);
  }
  void add_phi_pair_operand(Value *val, Value *pre_bb) {
    this->add_operand(val);
    this->add_operand(pre_bb);
  }
  virtual std::string print() override;

  };

  class Bitcast : public Instruction {
  public:
    Type *dest_ty_;
    Bitcast(OpID op, Value *val, Type *ty, BasicBlock *bb) : Instruction(ty, op, 1, bb), dest_ty_(ty) {
      set_operand(0, val);}
    virtual std::string print() override;
    
  
};

#endif 